/* ************************************************************************** 
 *             The Pebbles Personal Digital Assistant Project               *
 * **************************************************************************
 * This code was written as part of the Pebbles project at                  *
 * Carnegie Mellon University.  See the following for more information:     *
 *                http://www.cs.cmu.edu/~pebbles                            *
 * This code is being made available for free general non-commercial use.   *
 * For other conditions, see http://www.cs.cmu.edu/~pebbles/conditions.html *
 * Please contact us for more information: pebbles@cs.cmu.edu               *  
 *         Copyright (C) 1998-2000 - Carnegie Mellon University                  *
 * **************************************************************************/

/*
	This file describes the Pebbles protocol for communicating 
	between an application and PebblesPC using Windows Sockets

	These functions require wsock32.lib to be linked with a project

	Designed and implemented by Ben Bostwick
*/

#include "Pebbles.h"
#include "PebblesSocket.h"
#include "winsock.h"

#ifndef SD_BOTH
#define SD_BOTH 0x02
#endif

typedef struct _PebblesSocket {
	SOCKET socket;
	BOOL connected;
} PebblesSocket;

PebblesSocket pblSock;

/***********************************************************************
 *
 * FUNCTION:    PebblesSocketInit
 *
 * DESCRIPTION: Initializes Windows sockets and a PebblesSocket structure
 *              Call this function before any PebblesSocket functions are used
 *
 * PARAMETERS:  pbl	-- PebblesSocket to initialize
 *
 * RETURNED:    TRUE if Windows socket library is opened correctly, FALSE otherwise
 *
 ***********************************************************************/

bool PebblesSocketInit() {

	CHAR szError[100];                 // Error message string
	WSADATA WSAData;                    // Contains details of the Windows 
									  // Sockets implementation 
	// Initiate Windows Sockets.   
	if (WSAStartup (MAKEWORD(1,1), &WSAData) != 0)  {
		wsprintf (szError, TEXT("WSAStartup failed. Error: %d"), WSAGetLastError ());
		MessageBox (NULL, szError, TEXT("Error"), MB_OK);
		return FALSE;
	}

	pblSock.socket = INVALID_SOCKET;
	pblSock.connected = FALSE;

	return TRUE;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSocketClose
 *
 * DESCRIPTION: Closes Windows sockets and a PebblesSocket structure
 *              Call this function when PebblesSocket functions are no
 *              longer needed.  One PebblesSocketClose() call should occur
 *              for every PebblesSocketInit() call.
 *
 * PARAMETERS:  pbl	-- PebblesSocket to shutdown
 *
 * RETURNED:    nothing
 *
 ***********************************************************************/

void PebblesSocketClose() {
	shutdown (pblSock.socket, SD_BOTH);  // Close the socket for both reading and writing 
	closesocket (pblSock.socket);  
	WSACleanup ();					// windows sockets cleanup

	pblSock.socket = INVALID_SOCKET;
	pblSock.connected = FALSE;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSocketConnect
 *
 * DESCRIPTION: Connects to PebblesPC at specified hostname and port.
 *              The hostname should be an IP address in dotted format.
 *
 * PARAMETERS:  pbl			-- PebblesSocket used for connection
 *              hostname	-- Hostname of PC to connect to
 *              port		-- port number to connect to
 *
 * RETURNED:    TRUE if connection is established, FALSE otherwise
 *
 ***********************************************************************/

bool PebblesSocketConnect(char *plugname, char *hostname, int portnum) {
	int index = 0;                     // An integer 
	TCHAR szError[100];                 // Error message string 
	SOCKADDR_IN destination_sin;        // Server socket address 
	PHOSTENT phostent = NULL;           // Points to the HOSTENT structure 
									  // of the server 

	// Create a TCP/IP socket bound to the server.
	if ((pblSock.socket = socket (AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)  {
		wsprintf (szError, TEXT("Allocating socket failed. Error: %d"), WSAGetLastError ());
		MessageBox (NULL, szError, TEXT("Error"), MB_OK);
		return FALSE;
	}

	// Fill out the server socket address data. 
	destination_sin.sin_family = AF_INET;
	// Retrieve the host data corresponding to the host name.

/*  *BAM* This technique hangs if the PC isn't connected to 
********* a network, so use other technique instead.
    phostent = gethostbyname (hostname);
	if (phostent == NULL)   {
		wsprintf (szError, TEXT("Unable to get the host name. Error: %d"), WSAGetLastError ());
		MessageBox (NULL, szError, TEXT("Error"), MB_OK);
		closesocket (pblSock.socket);
		return FALSE;
	}

	// Assign the socket IP address. 
	memcpy ((char FAR *)&(destination_sin.sin_addr), phostent->h_addr, phostent->h_length);  // Convert to network ordering.
*/ /*BAM*/
/*BAM*/	unsigned long hl = inet_addr (hostname);
/*BAM*/ destination_sin.sin_addr.S_un.S_addr = hl;

	destination_sin.sin_port = htons (portnum);
  
	// Establish a connection to the server socket. 
	if (connect (pblSock.socket, (PSOCKADDR) &destination_sin, sizeof (destination_sin)) == SOCKET_ERROR)   {
		wsprintf (szError, TEXT("Connecting to the server failed. Error: %d"), WSAGetLastError ()); 
		MessageBox (NULL, szError, TEXT("Error"), MB_OK); 
		closesocket (pblSock.socket);
		return FALSE;
	}
	
	pblSock.connected = TRUE;

	// send the plugin name along
	int nameLen = strlen(plugname);
	int msgLen = nameLen + PEBBLES_SOCKET_HEADER_SIZE;
	char *msg = new char[msgLen];
	PEBBLES_SET_CMD(msg, CMD_SOCKET_PLUGIN_NAME);
	PEBBLES_SET_LEN(msg, nameLen);
	PEBBLES_SET_USERID(msg, PEBBLES_USERID_NO_USER);
	memcpy(PEBBLES_DATA(msg)+1, plugname, nameLen);
	PebblesSocketSend(msg, msgLen);

	return TRUE;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSocketReceive
 *
 * DESCRIPTION: Receives data waiting in the socket's receive buffer.
 *              The timeout is zero, so this function should be used
 *              after a call to PebblesSocketReceiveCheck() returns TRUE
 *
 *				Format of a socket message:
 *				1 byte			CMD
 *				2 bytes			LENGTH
 *				1 byte			USERID
 *				LENGTH bytes	DATA
 *
 * PARAMETERS:  pbl		   -- PebblesSocket to receive data from
 *              receiveBuf -- character buffer to receive the data
 *              receiveLen -- length of the receive buffer
 *
 * RETURNED:    TRUE if the data waiting to be read is put into the
 *              receiveBuf array.  FALSE if a socket error occurred.
 *              If receiveBuf is not large enough to hold all the received
 *              data, PebblesSocketReceive returns false and receiveLen is
 *              set to 0.  If this occurs, an application should loop until
 *              0 is not returned.
 *
 ***********************************************************************/

bool PebblesSocketReceive(char *receiveBuf, int *receiveLen) {
	int iReturn;
	//char *szReceive = new char[*receiveLen];
	int err;
	char szError[100];

	if(pblSock.connected) {
		// Receive data from the server socket with no timeout
		iReturn = recv (pblSock.socket, receiveBuf, *receiveLen, 0);

		*receiveLen = iReturn;
		// Verify if data is received.
		if (iReturn == SOCKET_ERROR)    {
			err = WSAGetLastError ();

			if(err == WSAEMSGSIZE) {
				*receiveLen = 0;
				return FALSE;
			}
			// there was an error.. most likely the socket closed
			pblSock.connected = FALSE;
			shutdown(pblSock.socket, SD_BOTH);
			closesocket (pblSock.socket);

			wsprintf (szError, TEXT("No data is received, recv failed."), TEXT(" Error: %d"), err);
			MessageBox (NULL, szError, TEXT("Client"), MB_OK); 
		}
		else if (iReturn == 0)    {
			// socket was closed, return
			*receiveLen = SOCKET_ERROR;
			pblSock.connected = FALSE;
			shutdown(pblSock.socket, SD_BOTH);
			closesocket (pblSock.socket);
		}
		else  {
			// everything is ok.
			return TRUE;
		}	
	}
	return FALSE;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSocketReceiveCheck
 *
 * DESCRIPTION: Checks to see if data is waiting at a PebblesSocket
 *
 * PARAMETERS:  pbl	-- PebblesSocket to check
 *
 * RETURNED:    TRUE if data is waiting at the socket, FALSE otherwise
 *
 ***********************************************************************/

bool PebblesSocketReceiveCheck() {
	if(pblSock.connected) {
		// create a set of sockets to check status
		FD_SET fd;
		FD_ZERO(&fd);
		// add our socket to the set
		FD_SET(pblSock.socket, &fd);

		// timeout is set to zero, so no waiting for data
		TIMEVAL tmval = {0, 0};

		// check socket
		select(0, &fd, NULL, NULL, &tmval); 

		// if the socket is in the returned set, data is waiting
		return (FD_ISSET(pblSock.socket, &fd) != 0);
	}
	// not connected
	return FALSE;
}

/***********************************************************************
 *
 * FUNCTION:    PebblesSocketSend
 *
 * DESCRIPTION: Send data through the PebblesSocket connection. Specify
 *              a Pebbles userid to send the message to on user, or 0
 *              in the userid field to send to all users.
 *
 * PARAMETERS:  pbl	 -- PebblesSocket to send
 *              data -- data to send
 *              len  -- total length of data to send
 *
 * RETURNED:    TRUE if data is successfully sent, FALSE otherwise
 *
 ***********************************************************************/

bool PebblesSocketSend(char *data, int len) {
	char szError[100];

	if(pblSock.connected) {
		if (send (pblSock.socket, data, len, 0) == SOCKET_ERROR)   { 
			wsprintf (szError, TEXT("Sending data to the server failed. Error: %d"), WSAGetLastError ());
			MessageBox (NULL, szError, TEXT("Error"), MB_OK);
			return FALSE;
		}
		return TRUE;
	}
	return FALSE;
}
